@font-face {
  font-family: 'Source Code Pro Custom', monospace;
  src: url(https://fonts.googleapis.com/css2?family=Source+Code+Pro:ital,wght@0,200..900;1,200..900&display=swap);
}

:root > * {
  --md-default-bg-color: #FFFFFF;
  --md-code-bg-color: #2E3440;
  --md-code-fg-color: #FFFFFF;
  --md-text-font-family: "Inter";
  --md-code-font: "Source Code Pro Custom";
  --md-typeset-a-color: #d16626; /*this is the brand color*/

  /* don't inherit white fg color for mermaid diagrams from --md-code-fg-color */
  --md-mermaid-label-fg-color: #000000;
  --md-mermaid-edge-color: #000000;
}

.index-pre-code {
  max-width: 700px;
  left: 50%;
}

.index-pre-code pre>code {
  text-align: left;
}

.md-clipboard::after {
  color: #FFFFFF;
  transition: color 0.3s ease-in-out;
}

.md-clipboard:hover::after {
  color: #D8DEE9;
}

.md-source-file {
  text-align: center;
  padding: 24px 0;
}

.md-typeset pre>code {
  border-radius: .2rem;
  box-shadow: 10px 5px 5px #D8DEE9;
}

.md-typeset p > code {
  background: #ECEFF4;
  color: #000000;
  font-weight: 500;
}

.md-typeset strong > code {
  background: #ECEFF4;
  color: #000000;
  font-weight: 500;
}

.md-content p > code {
  background: #ECEFF4;
  color: #000000;
  font-weight: 500;
}

.md-typeset td > code {
  background: #ECEFF4;
  color: #000000;
  font-weight: 500;
}

.md-typeset li > code {
  background: #ECEFF4;
  color: #000000;
  font-weight: 500;
}

.md-typeset code {
  font-weight: 500;
}

.md-typeset pre {
  margin-left: .5rem;
  margin-right: .5rem;
  margin-top: 2rem;
  margin-bottom: 2rem;
}

.language-python {
  background: #FFFFFF ! important
}

.language-bash {
  background: #FFFFFF ! important
}

.language-toml {
  background: #FFFFFF ! important
}

.language-text {
  background: #FFFFFF ! important
}

.language-json {
  background: #FFFFFF ! important
}

h1.title {
  color: #FFFFFF;
  margin: 0px 0px 5px;
}

h2.subtitle {
  margin: 5px 0px 25px;
  font-size: 1rem;
  max-width: 540px;
  margin: 0 auto;
}

.md-typeset {
  line-height: 24px;
  font-weight: 400;
}

.md-typeset h1 {
  font-weight: bold;
  color: #000000;
}

.md-typeset h2 {
  font-weight: bold;
  color: #000000;
}

span.md-ellipsis {
  color: black;
}

.md-nav__link--active {
  background-color: #ECEFF4;
}

.md-typeset ol li {
  margin-bottom: .1rem;
}

.md-typeset ul li {
  margin-bottom: .1rem;
}
